'From Squeak3.2gamma of 15 January 2002 [latest update: #4917] on 15 July 2002 at 8:06:05 pm'!
"Change Set:		SpreadPlugin-nk
Date:			15 July 2002
Author:			Ned Konz

This is a Plugin and supporting classes to allow Squeak to work with the Spread (http://www.spread.org) group communications system.

It is designed so that you can have background threads waiting for incoming messages without blocking all of Squeak.

I have only written the support code for Unix; it would be quite simple to port to other systems, especially if they have async file support like that in the Unix VM.

"!

Object subclass: #SpreadConnection
	instanceVariableNames: 'mbox privateGroupName privateName semaphore '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Project-SpreadPlugin'!

!SpreadConnection commentStamp: 'nk 7/15/2002 18:36' prior: 0!
I represent the client end of a single connection between a Spread client
and a Spread server.

I can receive messages sent to a group that I am a member of,
I can receive messages indicating changes to group membership,
and I can send messages to one or more groups.

I will block the current Process while waiting for the next message,
although a poll interface is available as well.!

Object subclass: #SpreadMessage
	instanceVariableNames: 'serviceType sender messageType numberOfGroups groups messageLength message errorCode endianMismatch '
	classVariableNames: 'ServiceTypeBits SpreadConstants '
	poolDictionaries: ''
	category: 'Project-SpreadPlugin'!

!SpreadMessage commentStamp: 'nk 7/15/2002 17:23' prior: 0!
This represents a raw message as received from SP_receive().

If you change the instance variable order you MUST change the corresponding receive primitive in SpreadPlugin.

instvars:

serviceType			various type flags for the message per Spread
sender				who sent the message (group or client)
messageType		user-defined message type, 16 bits, already endian corrected
numberOfGroups	how many groups in 'groups' (each 32 bytes)
groups				a buffer to hold the raw output of groups from Spread
messageLength		how many bytes can fit in 'message'
message				a buffer for the returned message
errorCode			the return from SP_receive. >=0 for bytes received, <0 for error.
endianMismatch		1 if there is an endian mismatch between machines, also used for signaling how long a too-short buffer needs to be.!

TestInterpreterPlugin subclass: #SpreadPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Project-SpreadPlugin'!

!SpreadPlugin commentStamp: 'nk 7/15/2002 18:41' prior: 0!
This is the Slang pluggable primitive source for the Spread Plugin.!


!Preferences class methodsFor: 'personalization' stamp: 'nk 7/14/2002 09:58'!
personalizeUserMenu: aMenu
	"The user has clicked on the morphic desktop with the yellow mouse button (option+click on the Mac); a menu is being constructed to present to the user in response; its default target is the current world.  In this method, you are invited to add items to the menu as per personal preferences.
	The default implementation, for illustrative purposes, sets the menu title to 'personal', and adds items for go-to-previous-project, show/hide flaps, and load code updates"
	
	aMenu addTitle: 'personal'.  "Remove or modify this as per personal choice"

	aMenu addStayUpItem.
	aMenu add: 'previous project' action: #goBack.
	aMenu add: 'load latest code updates' target: Utilities action: #updateFromServer.
	aMenu add: 'about this system...' target: Smalltalk action: #aboutThisSystem.
				
	aMenu addUpdating: #suppressFlapsString target: CurrentProjectRefactoring action: #currentToggleFlapsSuppressed.
	aMenu balloonTextForLastItem: 'Whether prevailing flaps should be shown in the project right now or not.'.

	aMenu add: 'vmmaker' target: VMMakerTool action: #openInWorld! !


!SpreadConnection methodsFor: 'primitives' stamp: 'nk 7/14/2002 16:22'!
primConnect: daemonName privateName: privName groupMembership: wantsMessages privateGroupBuf: groupNameBuf mbox: box semaIndex: semaIndex 
	<primitive: 'primitiveConnect' module:'SpreadPlugin'> "primitiveExternalCall" 
	^ SpreadMessage primitiveConnectFailure ! !

!SpreadConnection methodsFor: 'primitives' stamp: 'nk 7/15/2002 18:23'!
primDisconnect: conn 
	<primitive: 'primitiveDisconnect' module:'SpreadPlugin'> "primitiveExternalCall" 
	^ self primitiveFailed! !

!SpreadConnection methodsFor: 'primitives' stamp: 'nk 7/15/2002 18:24'!
primIsValid: conn
	<primitive: 'primitiveIsValid' module:'SpreadPlugin'> "primitiveExternalCall" 
	^ false! !

!SpreadConnection methodsFor: 'primitives' stamp: 'nk 7/15/2002 18:23'!
primJoin: conn withGroup: aGroup length: len
	<primitive: 'primitiveJoin' module:'SpreadPlugin'> "primitiveExternalCall" 
	^ self primitiveFailed! !

!SpreadConnection methodsFor: 'primitives' stamp: 'nk 7/15/2002 18:23'!
primLeave: conn fromGroup: aGroup length: len
	<primitive: 'primitiveLeave' module:'SpreadPlugin'> "primitiveExternalCall" 
	^ self primitiveFailed! !

!SpreadConnection methodsFor: 'primitives' stamp: 'nk 7/15/2002 18:23'!
primMulticast: conn serviceType: serviceType numGroups: numGroups groups: groups messType: messType messLen: messLen message: mess
	<primitive: 'primitiveMulticast' module:'SpreadPlugin'> "primitiveExternalCall" 
	^ self primitiveFailed! !

!SpreadConnection methodsFor: 'primitives' stamp: 'nk 7/15/2002 18:23'!
primPoll: conn
	<primitive: 'primitivePoll' module:'SpreadPlugin'> "primitiveExternalCall" 
	^ self primitiveFailed! !

!SpreadConnection methodsFor: 'primitives' stamp: 'nk 7/15/2002 18:23'!
primReceive: msg from: conn dropRecv: drop 
	<primitive: 'primitiveReceive' module:'SpreadPlugin'> "primitiveExternalCall" 
	^ self primitiveFailed ! !

!SpreadConnection methodsFor: 'Spread API' stamp: 'nk 7/15/2002 18:33'!
connectTo: daemonName privateName: privateNameOrNil wantGroupMembershipMessages: wantsGroupMembershipMessages 
	"Connect to the Spread daemon named by daemonName,
	with a (daemon-)unique privateName (if nil, one is assigned).
	wantsGroupMembershipMessages indicates whether or not group
	membership messages will be sent to me.
	Answer the Spread error code."

	| semaIndex groupNameBuf retval |
	semaphore _ Semaphore new.
	semaIndex _ Smalltalk registerExternalObject: semaphore.
	privateName _ privateNameOrNil.
	groupNameBuf _ ByteArray new: SpreadMessage maxGroupName.
	mbox _ ByteArray new: 12.	"must be sizeof(SpreadConnection)"
	retval _ self
				primConnect: daemonName asByteArray
				privateName: (privateName
						ifNil: ['']) asByteArray
				groupMembership: (wantsGroupMembershipMessages
						ifTrue: [1]
						ifFalse: [0])
				privateGroupBuf: groupNameBuf
				mbox: mbox
				semaIndex: semaIndex.
	retval ~= SpreadMessage acceptSession
		ifTrue: [Smalltalk unregisterExternalObject: semaphore.
			semaphore _ nil.
			mbox _ nil.
			^ retval].
	privateGroupName _ (groupNameBuf copyUpTo: 0) asString.
	^ retval! !

!SpreadConnection methodsFor: 'Spread API' stamp: 'nk 7/15/2002 18:33'!
disconnect
	"Disconnect from Spread.
	Wake up all Processes that were waiting
	for data to appear on this connection.
	Answer the Spread error code"

	| retval sem |
	mbox ifNil: [ ^-1 ].
	retval _ self primDisconnect: mbox.
	mbox _ nil.
	Smalltalk unregisterExternalObject: semaphore.
	sem _ semaphore.
	semaphore _ nil.
	privateGroupName _ nil.
	privateName _ nil.
	"Wake up all processes that were waiting for this connection"
	[ sem isEmpty ] whileFalse: [ sem signal ].
	^ retval! !

!SpreadConnection methodsFor: 'Spread API' stamp: 'nk 7/15/2002 18:33'!
isValid
	"Answer whether I am a valid connection.
	If I am not, all other Spread API calls will cause
	an Error"

	^self primIsValid: mbox! !

!SpreadConnection methodsFor: 'Spread API' stamp: 'nk 7/15/2002 18:33'!
join: aGroup
	"Join the group whose name is aGroup.
	Answer the Spread error code."

	^self primJoin: mbox withGroup: aGroup length: aGroup size! !

!SpreadConnection methodsFor: 'Spread API' stamp: 'nk 7/15/2002 18:33'!
leave: aGroup
	"Leave the group whose name is aGroup.
	Answer the Spread error code."

	^self primLeave: mbox fromGroup: aGroup length: aGroup size! !

!SpreadConnection methodsFor: 'Spread API' stamp: 'nk 7/15/2002 18:33'!
multicast: mess messageType: messType serviceType: serviceType groups: groups
	"Send the message 'mess', with user message type messType (16 bits),
	and Spread service type serviceType, to the groups whose names are listed
	in 'groups'.
	Answer the Spread error code."

	| groupsBuf |
	groupsBuf _ ByteArray new: (SpreadMessage maxGroupName * groups size).
	groups isEmpty ifFalse: [
		(0 to: groups size - 1) do: [ :n | | start str |
			start _ (n * SpreadMessage maxGroupName) + 1.
			str _ groups at: n + 1.
			groupsBuf replaceFrom: start
				to: start + str size - 1
				with: str asByteArray
				startingAt: 1.
		]
	].
	^self primMulticast: mbox
		serviceType: serviceType
		numGroups: groups size
		groups: groupsBuf
		messType: messType
		messLen: mess size
		message: mess! !

!SpreadConnection methodsFor: 'Spread API' stamp: 'nk 7/15/2002 18:33'!
poll
	"Answer the number of bytes waiting to be read,
	without blocking"

	^self primPoll: mbox! !

!SpreadConnection methodsFor: 'Spread API' stamp: 'nk 7/15/2002 18:34'!
receive
	"Block the current Process if necessary until data is ready,
	then answer a single message"

	[ self poll == 0 ] whileTrue: [ semaphore wait ].
	^self receiveOneMessage.! !

!SpreadConnection methodsFor: 'Spread API' stamp: 'nk 7/15/2002 18:34'!
receiveMessages
	"Block the current Process if necessary to wait for one or more messages,
	then answer them in a collection"

	| msgs |
	msgs _ OrderedCollection new.
	[ self poll == 0 ] whileTrue: [ semaphore wait ].
	[ self poll > 0 ] whileTrue: [
		msgs addLast: self receiveOneMessage.
	].
	^msgs.! !

!SpreadConnection methodsFor: 'private' stamp: 'nk 7/15/2002 18:26'!
receiveOneMessage
	"Receive a single message.
	Should only be used after waiting for the semaphore,
	as this will block all of Squeak if there is no data ready."
	| msg |
	mbox ifNil: [ ^nil ].
	msg _ SpreadMessage new.
	self
		primReceive: msg
		from: mbox
		dropRecv: 0.
	[ msg tooShort ] whileTrue: [
		msg reallocate.
		self
			primReceive: msg
			from: mbox
			dropRecv: 0.
	].
	^msg
! !


!SpreadMessage methodsFor: 'initialization'!
initialize
	serviceType _ 0.
	sender _ ByteArray new: self class maxGroupName.
	messageType _ 0.
	numberOfGroups _ self class defaultNumberOfGroups.
	groups _ ByteArray new: self class maxGroupName * numberOfGroups.
	messageLength _ self class defaultMessageLength.
	message _ ByteArray new: messageLength.
	errorCode _ 0.
	endianMismatch _ 0! !

!SpreadMessage methodsFor: 'error testing'!
bufferTooShort
	^ errorCode == self class bufferTooShort! !

!SpreadMessage methodsFor: 'error testing'!
groupsTooShort
	^ errorCode == self class groupsTooShort! !

!SpreadMessage methodsFor: 'error testing'!
tooShort
	^ self bufferTooShort
		or: [self groupsTooShort]! !

!SpreadMessage methodsFor: 'accessing' stamp: 'nk 7/15/2002 18:38'!
groups
	"Answer a collection of Strings naming the groups in this message"
	| start index |
	^ (0 to: numberOfGroups - 1)
		collect: [:n | 
			start _ n * self class maxGroupName + 1.
			index _ groups
						indexOf: 0
						startingAt: start
						ifAbsent: start.
			(String new: index - start)
				replaceFrom: 1
				to: index - start
				with: groups
				startingAt: start]! !

!SpreadMessage methodsFor: 'accessing' stamp: 'nk 7/15/2002 18:38'!
message
	"Answer a String with the message contents.
	Will answer an empty string if any error"

	errorCode <= 0 ifTrue: [ ^'' ].
	^ (String new: errorCode)
		replaceFrom: 1
		to: errorCode
		with: message
		startingAt: 1! !

!SpreadMessage methodsFor: 'accessing' stamp: 'nk 7/15/2002 18:38'!
messageType
	"Answer the user-defined message type (a 16-bit value)"

	^ messageType! !

!SpreadMessage methodsFor: 'accessing' stamp: 'nk 7/15/2002 18:39'!
printOn: strm
	"Print a representation of me on the given stream"
	strm nextPutAll: 'SpreadMessage('.
	self printServiceTypeOn: strm.
	strm nextPutAll: (errorCode < 0 ifTrue: [ ' err=' ] ifFalse: [ ' len=' ]).
	errorCode printOn: strm.
	strm nextPut: $)
! !

!SpreadMessage methodsFor: 'accessing' stamp: 'nk 7/15/2002 18:39'!
printServiceTypeOn: strm
	"print out a human-readable version of the service type"

	ServiceTypeBits do: [ :a |
		(serviceType anyMask: a second)
			ifTrue: [ strm nextPutAll: a first; nextPut: $, ].
	].
	strm position == 0 ifTrue: [
		^strm nextPutAll: 'UnknownServiceType'
	].
	strm position: strm position - 1.
! !

!SpreadMessage methodsFor: 'accessing' stamp: 'nk 7/15/2002 18:42'!
reallocate
	"Spread may have said that either or both of my message buffer
	or my group buffer were too short. Resize as needed so that a full
	reception can be made."

	self bufferTooShort
		ifTrue: [messageLength _ endianMismatch negated.
			message _ ByteArray new: messageLength].
	self groupsTooShort
		ifTrue: [numberOfGroups _ numberOfGroups negated.
			groups _ ByteArray new: self class maxGroupName * numberOfGroups].
! !

!SpreadMessage methodsFor: 'accessing' stamp: 'nk 7/15/2002 18:40'!
sender
	"Answer the name of the sender of the message (a group)"

	| index |
	index _ (sender
				indexOf: 0
				startingAt: 1
				ifAbsent: 1)
				- 1.
	^ (String new: index)
		replaceFrom: 1
		to: index
		with: sender
		startingAt: 1! !

!SpreadMessage methodsFor: 'accessing' stamp: 'nk 7/15/2002 18:40'!
serviceType
	"Answer a human-readable version of the service type"

	| strm |
	strm _ WriteStream on: (String new: 40).
	self printServiceTypeOn: strm.
	^strm contents! !

!SpreadMessage methodsFor: 'testing'!
hasEndianMismatch
	^ endianMismatch == 1! !

!SpreadMessage methodsFor: 'testing'!
hasError
	^ errorCode ~= 0! !

!SpreadMessage methodsFor: 'testing' stamp: 'nk 7/15/2002 12:18'!
isMembershipMessage
	^ (serviceType anyMask: self class membershipMessage)
		and: [ serviceType noMask: self class rejectMessage ]! !

!SpreadMessage methodsFor: 'testing' stamp: 'nk 7/15/2002 12:19'!
isRegularMessage
	^ (serviceType anyMask: self class regularMessage)
		and: [ serviceType noMask: self class rejectMessage ]
! !

!SpreadMessage methodsFor: 'testing' stamp: 'nk 7/15/2002 12:19'!
isRejectMessage
	^ serviceType anyMask: self class rejectMessage! !


!SpreadMessage class methodsFor: 'Spread constants'!
acceptSession
	^ SpreadConstants at: #AcceptSession! !

!SpreadMessage class methodsFor: 'Spread constants'!
bufferTooShort
	^ SpreadConstants at: #BufferTooShort! !

!SpreadMessage class methodsFor: 'Spread constants'!
groupsTooShort
	^ SpreadConstants at: #GroupsTooShort! !

!SpreadMessage class methodsFor: 'Spread constants'!
maxGroupName
	^ SpreadConstants at: #MaxGroupName! !

!SpreadMessage class methodsFor: 'Spread constants' stamp: 'nk 7/15/2002 17:57'!
membershipMessage
	^SpreadConstants at: #MembershipMessage! !

!SpreadMessage class methodsFor: 'Spread constants' stamp: 'nk 7/14/2002 16:21'!
primitiveConnectFailure
	^0! !

!SpreadMessage class methodsFor: 'Spread constants' stamp: 'nk 7/15/2002 17:57'!
regularMessage
	^SpreadConstants at: #RegularMessage! !

!SpreadMessage class methodsFor: 'Spread constants' stamp: 'nk 7/15/2002 17:57'!
rejectMessage
	^SpreadConstants at: #RejectMessage! !

!SpreadMessage class methodsFor: 'class initialization' stamp: 'nk 7/15/2002 17:31'!
initialize
	"SpreadMessage initialize."
	self initializeServiceTypeTable.
	self initializeConstants.! !

!SpreadMessage class methodsFor: 'class initialization' stamp: 'nk 7/15/2002 17:59'!
initializeConstants
	"SpreadMessage initializeServiceTypeTable; initializeConstants."
	SpreadConstants _ IdentityDictionary new.
	ServiceTypeBits do: [:t1 | SpreadConstants at: t1 first put: t1 second].
	{
#(#LowPriority 0 ). #(#MediumPriority 1 ). #(#HighPriority 2 ). 
#(#DefaultSpreadPort 4803 ). 
{ #SpreadVersion. ((3 bitShift: 24) bitOr: (16 bitShift: 16)) }. 
#(#MaxGroupName 32 ). #(#MaxPrivateName 10 ). #(#MaxProcName 20 ). #(#MaxClientScatterElements 100 ).
#(#SelfDiscard 64 ). 
#(#DropRecv 16r1000000). 
#(#EndianReserved 16r80000080 ). #(#Reserved 16r003fc000). #(#RejectMess 16r400000).
#(#AcceptSession 1 ).
#(#IllegalSpread -1 ). #(#CouldNotConnect -2 ). #(#RejectQuota -3 ). #(#RejectNoName -4 ). #(#RejectIllegalName -5 ). #(#RejectNotUnique -6 ). #(#RejectVersion -7 ). #(#ConnectionClosed -8 ). #(#RejectAuth -9 ). #(#IllegalSession -11 ). #(#IllegalService -12 ). #(#IllegalMessage -13 ). #(#IllegalGroup -14 ). #(#BufferTooShort -15 ). #(#GroupsTooShort -16 ). #(#MessageTooLong -17 ). 
}
		do: [:t1 | SpreadConstants at: t1 first put: t1 second]! !

!SpreadMessage class methodsFor: 'class initialization' stamp: 'nk 7/15/2002 17:58'!
initializeServiceTypeTable
	"decoder for the serviceType field."
	"SpreadMessage initializeServiceTypeTable"
	ServiceTypeBits _ {
		#(#RegularMessage 63 ).
		#(#Unreliable 1 ).
		#(#Reliable 2 ).
		#(#FIFO 4 ).
		#(#Causal 8 ).
		#(#Agreed 16 ).
		#(#Safe 32 ).
		#(#SelfDiscard 64 ).

		#(#MembershipMessage 16r3F00 ).
		#(#Regular 16r1000 ).
		#(#Transition 16r2000 ).
		#(#CausedByJoin 16r100 ).
		#(#CausedByLeave 16r200 ).
		#(#CausedByDisconnect 16r400 ).
		#(#CausedByNetwork 16r800 ).

		#(#RejectMessage 16r400000).
	}.! !

!SpreadMessage class methodsFor: 'defaults' stamp: 'nk 7/15/2002 18:14'!
defaultMessageLength
	^ 10! !

!SpreadMessage class methodsFor: 'defaults' stamp: 'nk 7/15/2002 18:00'!
defaultNumberOfGroups
	^ 1! !

!SpreadMessage class methodsFor: 'instance creation' stamp: 'nk 7/15/2002 09:18'!
new
	^super new initialize! !


!SpreadPlugin methodsFor: 'primitives'!
initialiseModule
	self export: true.
	^ self
		cCode: 'sqSpreadInit()'
		inSmalltalk: [true]! !

!SpreadPlugin methodsFor: 'primitives' stamp: 'nk 7/15/2002 08:31'!
primitiveConnect: daemonName privateName: privateName groupMembership: wantsGroupMsgs privateGroupBuf: groupBuf mbox: mbox semaIndex: semaIndex 
	| daemonNameSize privateNameSize retval s |
	self primitive: 'primitiveConnect' parameters: #(#String #String #SmallInteger #ByteArray #Oop #SmallInteger ).
	self var: #s type: 'SpreadConnection *'.
	daemonNameSize _ interpreterProxy stSizeOf: (daemonName - 4).
	privateNameSize _ interpreterProxy stSizeOf: (privateName - 4).
	s _ self spreadConnectionValueOf: mbox.
	interpreterProxy failed ifFalse: [
		self cCode: 'retval = sqSpreadConnect(s, daemonName, daemonNameSize, privateName, privateNameSize, wantsGroupMsgs, groupBuf, semaIndex)'.
		"interpreterProxy success: retval not."
	].
	^ retval asSmallIntegerObj ! !

!SpreadPlugin methodsFor: 'primitives' stamp: 'nk 7/15/2002 07:26'!
primitiveDisconnect: mbox 
	| retval s |
	self primitive: 'primitiveDisconnect' parameters: #(#Oop ).
	self var: #s type: 'SpreadConnection *'.
	s _ self spreadConnectionValueOf: mbox.
	interpreterProxy failed ifTrue: [ ^nil ].
	self cCode: 'retval = sqSpreadDisconnect(s)'.
	^ retval asSmallIntegerObj ! !

!SpreadPlugin methodsFor: 'primitives' stamp: 'nk 7/15/2002 16:46'!
primitiveIsValid: mbox
	| retval s |
	self primitive: 'primitiveIsValid' parameters: #(#Oop).
	self var: #s type: 'SpreadConnection *'.
	s _ self spreadConnectionValueOf: mbox.
	interpreterProxy failed ifTrue: [ ^nil ].
	self cCode: 'retval = sqSpreadIsValid(s)'.
	retval == 1 ifTrue: [ ^interpreterProxy trueObject ].
	^interpreterProxy falseObject
! !

!SpreadPlugin methodsFor: 'primitives' stamp: 'nk 7/15/2002 16:39'!
primitiveJoin: mbox withGroup: group length: len
	| retval s |
	self primitive: 'primitiveJoin' parameters: #(#Oop #String #SmallInteger).
	self var: #s type: 'SpreadConnection *'.
	s _ self spreadConnectionValueOf: mbox.
	interpreterProxy failed ifTrue: [ ^nil ].
	self cCode: 'retval = sqSpreadJoin(s, group, len)'.
	^ retval asSmallIntegerObj ! !

!SpreadPlugin methodsFor: 'primitives' stamp: 'nk 7/15/2002 16:39'!
primitiveLeave: mbox fromGroup: group length: len
	| retval s |
	self primitive: 'primitiveLeave' parameters: #(#Oop #String #SmallInteger).
	self var: #s type: 'SpreadConnection *'.
	s _ self spreadConnectionValueOf: mbox.
	interpreterProxy failed ifTrue: [ ^nil ].
	self cCode: 'retval = sqSpreadLeave(s, group, len)'.
	^ retval asSmallIntegerObj ! !

!SpreadPlugin methodsFor: 'primitives' stamp: 'nk 7/15/2002 12:07'!
primitiveMulticast: mbox serviceType: serviceType numGroups: numGroups groups: groups messType: messType messLen: messLen message: mess
	| retval s |
	self primitive: 'primitiveMulticast' parameters: #(#Oop #SmallInteger #SmallInteger #ByteArray #SmallInteger #SmallInteger #ByteArray ).
	self var: #s type: 'SpreadConnection *'.
	s _ self spreadConnectionValueOf: mbox.
	interpreterProxy failed ifTrue: [ ^nil ].
	self cCode: 'retval = sqSpreadMultigroupMulticast(s, serviceType, numGroups, groups, messType, messLen, mess)'.
	^ retval asSmallIntegerObj ! !

!SpreadPlugin methodsFor: 'primitives' stamp: 'nk 7/15/2002 16:39'!
primitivePoll: mbox
	| retval s |
	self primitive: 'primitivePoll' parameters: #(#Oop).
	self var: #s type: 'SpreadConnection *'.
	s _ self spreadConnectionValueOf: mbox.
	interpreterProxy failed ifTrue: [ ^nil ].
	self cCode: 'retval = sqSpreadPoll(s)'.
	^ retval asSmallIntegerObj ! !

!SpreadPlugin methodsFor: 'primitives' stamp: 'nk 7/15/2002 18:10'!
primitiveReceive: msg from: mbox dropRecv: drop
	| serviceType sender messageType numberOfGroups groups messageLength message errorCode endianMismatch s |
	self var: #sender type: 'char *'.
	self var: #groups type: 'char *'.
	self var: #message type: 'char *'.
	self var: #s type: 'SpreadConnection *'.
	self primitive: 'primitiveReceive' parameters: #(#Oop #Oop #SmallInteger ).
	"0=serviceType 1=sender 2=messageType 3=numberOfGroups 4=groups
		5=messageLength 6=message 7=errorCode 8=endianMismatch"
	sender _ interpreterProxy fetchArray: 1 ofObject: msg.
	numberOfGroups _ interpreterProxy fetchInteger: 3 ofObject: msg.
	groups _ interpreterProxy fetchArray: 4 ofObject: msg.
	messageLength _ interpreterProxy fetchInteger: 5 ofObject: msg.
	message _ interpreterProxy fetchArray: 6 ofObject: msg.
	s _ self spreadConnectionValueOf: mbox.
	interpreterProxy failed ifTrue: [ ^nil ].
	"TODO: make this work with big-endian machines!! (remove messageType cast)"
	self cCode: 'errorCode = sqSpreadReceive(s, &serviceType, sender,
		numberOfGroups, &numberOfGroups, groups,
		(int16*)&messageType, &endianMismatch,
		messageLength, message, drop)'.
	interpreterProxy storeInteger: 7 ofObject: msg withValue: errorCode.
	interpreterProxy storeInteger: 0 ofObject: msg withValue: serviceType.
	interpreterProxy storeInteger: 3 ofObject: msg withValue: numberOfGroups.
	interpreterProxy storeInteger: 2 ofObject: msg withValue: messageType.
	interpreterProxy storeInteger: 8 ofObject: msg withValue: endianMismatch.
	^errorCode asSmallIntegerObj! !

!SpreadPlugin methodsFor: 'primitives'!
shutdownModule
	self export: true.
	^ self
		cCode: 'sqSpreadShutdown()'
		inSmalltalk: [true]! !

!SpreadPlugin methodsFor: 'primitives' stamp: 'nk 7/15/2002 06:19'!
spreadConnectionValueOf: oop
	"Return a pointer to the first byte of the Spread connection record
	within the given Smalltalk bytes object, or nil if oop is not an Spread connection record."

	self returnTypeC: 'SpreadConnection *'.
	interpreterProxy success:
		((interpreterProxy isIntegerObject: oop) not and:
		 [(interpreterProxy isBytes: oop) and:
		 [(interpreterProxy slotSizeOf: oop) = (self cCode: 'sizeof(SpreadConnection)')]]).
	interpreterProxy failed ifTrue: [^ nil].
	^ self cCode: '(SpreadConnection *) (oop + 4)'
! !


!SpreadPlugin class methodsFor: 'as yet unclassified'!
hasHeaderFile
	^ true! !

SpreadPlugin removeSelector: #breakpoint!
SpreadPlugin removeSelector: #primJoin:withGroup:length:!
SpreadPlugin removeSelector: #primLeave:fromGroup:length:!
SpreadPlugin removeSelector: #primPoll:!
SpreadPlugin removeSelector: #primitiveConnect:name:privateName:groupMembership:privateGroupBuf:mbox:semaIndex:!
SpreadPlugin removeSelector: #primitiveConnect:name:privateName:groupMembership:privateGroupBuf:semaIndex:!

!SpreadPlugin reorganize!
('primitives' initialiseModule primitiveConnect:privateName:groupMembership:privateGroupBuf:mbox:semaIndex: primitiveDisconnect: primitiveIsValid: primitiveJoin:withGroup:length: primitiveLeave:fromGroup:length: primitiveMulticast:serviceType:numGroups:groups:messType:messLen:message: primitivePoll: primitiveReceive:from:dropRecv: shutdownModule spreadConnectionValueOf:)
!

SpreadMessage initialize!
SpreadMessage removeSelector: #errorCode:!
SpreadConnection removeSelector: #connectTo:name:privateName:wantGroupMembershipMessages:!
SpreadConnection removeSelector: #primConnect:name:privateName:groupMembership:privateGroupBuf:mbox:semaIndex:!
SpreadConnection removeSelector: #primConnect:name:privateName:groupMembership:privateGroupBuf:semaIndex:!
SpreadConnection removeSelector: #primJoin:length:!
SpreadConnection removeSelector: #primLeave:length:!

!SpreadConnection reorganize!
('primitives' primConnect:privateName:groupMembership:privateGroupBuf:mbox:semaIndex: primDisconnect: primIsValid: primJoin:withGroup:length: primLeave:fromGroup:length: primMulticast:serviceType:numGroups:groups:messType:messLen:message: primPoll: primReceive:from:dropRecv:)
('Spread API' connectTo:privateName:wantGroupMembershipMessages: disconnect isValid join: leave: multicast:messageType:serviceType:groups: poll receive receiveMessages)
('private' receiveOneMessage)
!

